#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>

#define NUMCHILD 3			// Numero di processi figli generati

volatile int sig_counter=NUMCHILD;	// Contatore dei processi figli

//-----------------------------------------------------------------------------
// Interrupt handler per SIGCHLD
// Stampa un messaggio e decrementa il contatore dei processi figli
// N.B. Usa 'write', non 'printf'!

void clean_up_child_process(int sig_number)
{
  char msg[]="Ricevuto segnale SIGCHLD!\n";
  write(1,msg,strlen(msg));
  sig_counter-=1;
}

//-----------------------------------------------------------------------------
// Funzione principale
// 1- inizializza il signal handler per SIGCHLD
// 2- genera NUMCHLD processi figli, ognuno dei quali esce con rc diverso
// 3- cicla indefinitamente in attesa di ricevere tutti i SIGCHLD
// 4- stampa i rc di tutti i processi figli
// N.B. Usa 'signal', non 'sigaction' per ragioni di semplicita'

int main()
{
  int i;				// contatore generico
  int rc;				// codice di ritorno generico
  int status;				// stato del processo figlio 
  struct sigaction newaction, oldaction;// azioni da associare a SIGCHLD

// installazione del signal handler per SIGCHLD

  memset(&newaction,0,sizeof(newaction));
  newaction.sa_handler=&clean_up_child_process;
  rc=sigaction(SIGCHLD,&newaction,&oldaction);	
  if (rc != 0) {
    printf("Non posso installare il signal handler!\n");
    exit(-1);
    }

// ogni processo figlio ritorna il proprio indice ed esce

  for (i=0;i<NUMCHILD;i++) {
    if (fork()==0) {
      printf("Figlio generato correttamente...\n");
      exit(i);				
      }
    }

// il padre cicla indefinitamente finche' non riceve tutti i SIGCHLD

  while (1) {if (sig_counter==0) break;}; 

  printf("Ricevuti tutti i segnali SIGCHLD!\n");

// il padre cicla sui figli per recuperare l'exit code

  for (i=0;i<NUMCHILD;i++) {
    wait(&status);
    printf("Figlio terminato con stato %d.\n",status);
    }

  exit(0);
}
